#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <threads.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <unistd.h>

struct thread_args
{
    mtx_t *counts_mtx;
    int *counts;
    int *should_exit;
};

void* worker(void *arg)
{
    struct thread_args *args = (struct thread_args*) arg;
    int num;
    while (! *(args->should_exit)) {
        num = random() % 10;
        mtx_lock(&(args->counts_mtx[num]));
        args->counts[num]++;
        mtx_unlock(&(args->counts_mtx[num]));
    }
    printf("worker exit\n");
    return NULL;
}

void print_stat(struct thread_args *args)
{
    for (int i = 0; i < 10; i++) {
        mtx_lock(&(args->counts_mtx[i]));
        printf("%d: %d\n", i, args->counts[i]);
        mtx_unlock(&(args->counts_mtx[i]));
    }
    printf("====\n");
}

void* show_stat(void *arg)
{
    struct thread_args *args = (struct thread_args*) arg;
    while (! *(args->should_exit)) {
        print_stat(args);
        sleep(1);
    }
    printf("show_stat exit\n");
    return NULL;
}

void* check_result(void *arg)
{
    const int N = 5000000;
    struct thread_args *args = (struct thread_args*) arg;
    while (! *(args->should_exit)) {
        mtx_lock(&(args->counts_mtx[1]));
        if (args->counts[1] >= N) {
            *(args->should_exit) = 1;
            mtx_unlock(&(args->counts_mtx[1]));
            break;
        }
        mtx_unlock(&(args->counts_mtx[1]));
        usleep(200);
    }
    printf("counts[1]=%d\n", args->counts[1]);
    printf("got %d 1s, final stats:\n", N);
    print_stat(args);
    return NULL;
}

int main()
{
    struct timeval start_time;
    struct timeval end_time;

    gettimeofday(&start_time, NULL);
    srandom(start_time.tv_sec);

    mtx_t counts_mtx[10];
    int counts[10] = {0};

    for (int i = 0; i < 10; i++) {
        int r = mtx_init(&counts_mtx[i], mtx_plain);
        if (r != thrd_success) {
            printf("Error: create mutex failed.\n");
            for (int j = 0; j < i; j++) {
                mtx_destroy(&counts_mtx[j]);
            }
            return -1;
        }
    }
    int should_exit = 0;

    struct thread_args arg;
    arg.counts_mtx = counts_mtx;
    arg.counts = counts;
    arg.should_exit = &should_exit;

    pthread_t worker1;
    pthread_create(&worker1, NULL, worker, &arg);
    pthread_t worker2;
    pthread_create(&worker2, NULL, worker, &arg);
    pthread_t stat;
    pthread_create(&stat, NULL, show_stat, &arg);
    pthread_t result;
    pthread_create(&result, NULL, check_result, &arg);

    pthread_join(result, NULL);
    gettimeofday(&end_time, NULL);
    printf("total time: %.2f seconds\n",
            (end_time.tv_sec - start_time.tv_sec) +
            (end_time.tv_usec - start_time.tv_usec) / 1000000.0);
    pthread_join(worker1, NULL);
    pthread_join(worker2, NULL);
    pthread_join(stat, NULL);

    for (int i = 0; i < 10; i++) {
        mtx_destroy(&counts_mtx[i]);
    }
    return 0;
}
