import time
import threading
import random


def worker(lock, counts, should_exit):
    while True:
        if should_exit:
            break
        with lock:
            counts[random.randint(0, 9)] += 1
    print("worker exit")


def print_stat(lock, counts):
    with lock:
        for i, c in enumerate(counts):
            print("%d: %d" % (i, c))
    print("====")


def show_stat(lock, counts, should_exit):
    while True:
        if should_exit:
            break
        print_stat(lock, counts)
        time.sleep(1)
    print("show_stat exit")


def check_result(lock, counts, should_exit):
    #N = 5000000
    N = 500000
    while True:
        if should_exit:
            break
        with lock:
            if counts[1] >= N:
                should_exit.append(1)
                break
        time.sleep(0.2)
    print("counts[1]=%d" % (counts[1],))
    print("got %d 1s, final stats:" % (N,))
    print_stat(lock, counts)


def main():
    start = time.time()
    random.seed()
    lock = threading.Lock()
    counts = [0] * 10
    should_exit = []
    worker1 = threading.Thread(target=worker, args=(lock, counts, should_exit))
    worker1.start()
    worker2 = threading.Thread(target=worker, args=(lock, counts, should_exit))
    worker2.start()
    stat = threading.Thread(target=show_stat, args=(lock, counts, should_exit))
    stat.start()
    result = threading.Thread(target=check_result, args=(lock, counts, should_exit))
    result.start()
    
    result.join()
    end = time.time()
    print("total time: %.2f seconds" % (end - start,))
    worker1.join()
    worker2.join()
    stat.join()

if __name__ == "__main__":
    main()
